// Package test provides helpers to test the various pion transports implementations.
// The tests are standardized around the io.ReadWriteCloser interface.
// This package is meant to be used in addition to golang.org/x/net/nettest.
package test
